from .backend import core
from .helper import FlowText


class Window:
    def __init__(self, app, x, y, w, h):
        self.app = app
        self.core = core.Window(app.core, x, y, w, h)
        self.text = FlowText(w - 2, h - 2, self.on_flow)

        self.app.windows.append(self)

    @property
    def x(self):
        return self.core.x

    @property
    def y(self):
        return self.core.y

    @property
    def w(self):
        return self.core.w

    @property
    def h(self):
        return self.core.h

    def move_to(self, x, y):
        self.core.set_pos(x, y)

    def move_by(self, dx, dy):
        x = self.x + dx
        if x < 0: x = 0
        elif x + self.w > self.app.w: x = self.app.w - self.w

        y = self.y + dy
        if y < 0: y = 0
        elif y + self.h > self.app.h: y = self.app.h - self.h

        self.move_to(x, y)

    def size_to(self, w, h):
        self.core.set_size(w, h)
        self.text.resize(w - 2, h - 2)

    def size_by(self, dw, dh):
        w = self.w + dw
        if w < 2: w = 2
        elif self.x + w > self.app.w: w = self.app.w - self.x

        h = self.h + dh
        if h < 2: h = 2
        elif self.y + h > self.app.h: h = self.app.h - self.y

        self.size_to(w, h)

    def on_flow(self, flow):
        assert flow == self.text
        view = flow.view()
        if view is None: return
        self.core.write_all(1, 1, view)




import os

if os.name == 'nt':
    # taken from https://stackoverflow.com/questions/8677627/getting-mouse-presses-on-a-console-window-for-python

    import win32con
    import win32file
    from win32console import *
    import traceback, time

    virtual_keys={}
    for k,v in win32con.__dict__.items():
        if k.startswith('VK_'):
            virtual_keys[v]=k 

    free_console=True
    try:
        AllocConsole()
    except error as exc:
        if exc.winerror!=5:
            raise
        ## only free console if one was created successfully
        free_console=False

    stdout=GetStdHandle(STD_OUTPUT_HANDLE)

    conin=PyConsoleScreenBufferType( win32file.CreateFile( "CONIN$", win32con.GENERIC_READ|win32con.GENERIC_WRITE, win32con.FILE_SHARE_READ, None, win32con.OPEN_EXISTING, 0, 0))

    conin.SetConsoleMode(ENABLE_WINDOW_INPUT | ENABLE_MOUSE_INPUT)

    newbuffer=CreateConsoleScreenBuffer()
    newbuffer.SetConsoleActiveScreenBuffer()
    newbuffer.SetConsoleTextAttribute(FOREGROUND_RED|FOREGROUND_INTENSITY
            |BACKGROUND_GREEN|BACKGROUND_INTENSITY)
    newbuffer.WriteConsole('This is a new screen buffer\n')

    newbuffer.SetConsoleTextAttribute(FOREGROUND_RED|FOREGROUND_INTENSITY
            |BACKGROUND_GREEN|BACKGROUND_INTENSITY)
    newbuffer.WriteConsole('Press some keys, click some characters with the mouse\n')

    newbuffer.SetConsoleTextAttribute(FOREGROUND_BLUE|FOREGROUND_INTENSITY
            |BACKGROUND_RED|BACKGROUND_INTENSITY)
    newbuffer.WriteConsole('Hit "Esc" key to quit\n')


    breakout=False
    while not breakout:
        input_records=conin.ReadConsoleInput(10)
        for input_record in input_records:
            if input_record.EventType==KEY_EVENT:
                if input_record.KeyDown:
                    if input_record.Char=='\0':
                        newbuffer.WriteConsole(virtual_keys.get(input_record.VirtualKeyCode, 'VirtualKeyCode: %s' %input_record.VirtualKeyCode))
                    else:
                        newbuffer.WriteConsole(input_record.Char)
                    if input_record.VirtualKeyCode==win32con.VK_ESCAPE:
                        breakout=True
                        break
            elif input_record.EventType==MOUSE_EVENT:
                if input_record.EventFlags==0:  ## 0 indicates a button event
                    if input_record.ButtonState!=0:   ## exclude button releases
                        pos=input_record.MousePosition
                        # switch the foreground and background colors of the character that was clicked
                        attr=newbuffer.ReadConsoleOutputAttribute(Length=1, ReadCoord=pos)[0]
                        new_attr=attr
                        if attr&FOREGROUND_BLUE:
                            new_attr=(new_attr&~FOREGROUND_BLUE)|BACKGROUND_BLUE
                        if attr&FOREGROUND_RED:
                            new_attr=(new_attr&~FOREGROUND_RED)|BACKGROUND_RED
                        if attr&FOREGROUND_GREEN:
                            new_attr=(new_attr&~FOREGROUND_GREEN)|BACKGROUND_GREEN

                        if attr&BACKGROUND_BLUE:
                            new_attr=(new_attr&~BACKGROUND_BLUE)|FOREGROUND_BLUE
                        if attr&BACKGROUND_RED:
                            new_attr=(new_attr&~BACKGROUND_RED)|FOREGROUND_RED
                        if attr&BACKGROUND_GREEN:
                            new_attr=(new_attr&~BACKGROUND_GREEN)|FOREGROUND_GREEN
                        newbuffer.WriteConsoleOutputAttribute((new_attr,),pos)
            else:
                newbuffer.WriteConsole(str(input_record))
        time.sleep(0.1)

    newbuffer.Close()

    if free_console:
         FreeConsole()

else:
    # taken from https://stackoverflow.com/questions/56303971/how-to-enable-mouse-movement-events-in-python-curses

    import curses

    screen = curses.initscr()
    screen.keypad(1)
    curses.curs_set(0)
    curses.mousemask(curses.ALL_MOUSE_EVENTS | curses.REPORT_MOUSE_POSITION)
    curses.flushinp()
    curses.noecho()
    screen.clear()

    while True:
        key = screen.getch()
        screen.clear()
        screen.addstr(0, 0, 'key: {}'.format(key))
        if key == curses.KEY_MOUSE:
            _, x, y, _, button = curses.getmouse()
            screen.addstr(1, 0, f'x, y, button = {x}, {y}, {button}')
        elif key == 27:
            break

    curses.endwin()
    curses.flushinp()




import curses


debug = None


def main(screen):
    global debug
    curses.flushinp()
    curses.set_escdelay(100)
    curses.mousemask(curses.ALL_MOUSE_EVENTS | curses.REPORT_MOUSE_POSITION)

    win = curses.newwin(5, 30, 5, 30)
    win.keypad(True)
    win.box()
    win.addstr(1, 1, 'hold click and drag me')
    win.addstr(2, 1, 'ESC to quit')

    moving = None

    while True:
        key = win.getch()
        if key == 27: # escape
            break

        elif key == curses.KEY_MOUSE:
            _, x, y, _, btn = curses.getmouse()

            if btn & curses.BUTTON1_PRESSED:
                wy, wx = win.getbegyx()
                h, w = win.getmaxyx()

                if x >= wx and x < wx + w and y >= wy and y < wy + h:
                    moving = (x, y)

            elif moving:
                sx = moving[0]
                sy = moving[1]

                if x != sx or y != sy:
                    dx = x - sx
                    dy = y - sy
                    wy, wx = win.getbegyx()
                    h, w = win.getmaxyx()

                    nx = wx + dx
                    if nx < 0: nx = 0
                    elif nx + w >= curses.COLS: nx = curses.COLS - w

                    ny = wy + dy
                    if ny < 0: ny = 0
                    elif ny + h >= curses.LINES: ny = curses.LINES - h

                    screen.clear()
                    screen.refresh()
                    win.mvwin(ny, nx)
                    moving = (x, y)

                if btn & curses.BUTTON1_RELEASED:
                    moving = None


curses.wrapper(main)
curses.flushinp()

if debug is not None:
    print(f'debug: {debug}')






class FlowText:
    def __init__(self, w, h, notify=None, words=False, align=-1):
        assert w >= 0 and h >= 0
        self.w = w
        self.h = h
        self.notify = notify
        self.words = words
        self.align = align

        self.prints = [ ]
        self.enter = True

        self.lines = [ ]
        self.scroll = -1
        self.flow = (0, 0)

    def resize(self, w, h):
        assert w >= 0 and h >= 0
        if w == self.w and h == self.h: return
        self.h = h
        if w != self.w:
            self.w = w
            self.reflow(True)
        else:
            if self.notify: self.notify(self)

    def word_wrap(self, words=None):
        if words is None:
            self.words = not self.words
            self.reflow(True)

        elif words != self.words:
            self.words = words
            self.reflow(True)

    def left(self):
        if self.align == -1: return
        self.align = -1
        if self.notify: self.notify(self)

    def center(self):
        if self.align == 0: return
        self.align = 0
        if self.notify: self.notify(self)

    def right(self):
        if self.align == 1: return
        self.align = 1
        if self.notify: self.notify(self)

    def top(self):
        if self.scroll == 0: return
        self.scroll = 0
        if self.notify: self.notify(self)

    def bottom(self):
        if self.scroll < 0: return
        self.scroll = -1
        if self.notify: self.notify(self)

    def up(self, lines=1):
        assert lines > 0
        if self.scroll == 0: return
        diff = len(self.lines) - self.h
        if diff > 0:
            self.scroll -= lines
            if self.scroll > diff:
                self.scroll = diff
            elif self.scroll < 0:
                self.scroll = 0
        else:
            self.scroll = 0
        if self.notify: self.notify(self)

    def down(self, lines=1):
        assert lines > 0
        if self.scroll < 0: return
        diff = len(self.lines) - self.h
        if diff > 0:
            self.scroll += lines
            if self.scroll > diff:
                self.scroll = diff
        else:
            self.scroll = -1
        if self.notify: self.notify(self)

    def print(self, txt='', enter=True):
        if not isinstance(txt, str):
            txt = f'{txt}'

        prints = [ ]

        while '\n' in txt:
            idx = txt.index('\n')
            prints.append(txt[:idx])
            txt = txt[idx+1:]

        prints.append(txt)

        if self.enter:
            self.prints.extend(prints)

        else:
            self.prints[-1] += prints[0]
            self.prints.extend(prints[1:])

        self.enter = enter
        self.reflow()

    def reflow(self, reset=False):
        if reset:
            self.flow = (0, 0)

        if self.w < 1 or self.h < 1:
            return

        if self.flow[0] and self.flow[1]:
            self.lines = self.lines[:self.flow[1]]

        elif self.lines:
            self.lines = [ ]

        for entry in self.prints[self.flow[0]:]:
            while len(entry) > self.w:
                chunk = entry[:self.w]

                if self.words and entry[self.w] != ' ' and chunk[-1] != ' ' \
                        and ' ' in chunk:
                    idx = chunk.rindex(' ')
                    chunk = entry[:idx]
                    entry = entry[idx+1:]

                else:
                    entry = entry[self.w:]

                self.lines.append(chunk)

            self.lines.append(entry)

        if self.enter:
            self.flow = (len(self.prints), len(self.lines))

        else:
            self.flow = (len(self.prints) - 1, len(self.lines) - 1)

        if self.notify:
            self.notify(self)

    def view(self):
        if self.w < 1 or self.h < 1:
            return None

        if self.scroll < 0:
            showing = self.lines[-self.h:]

        else:
            showing = self.lines[self.scroll:self.scroll+self.h]

        while len(showing) < self.h:
            showing.append('')

        if self.align < 0: # left
            return [ line + ' ' * (self.w - len(line)) for line in showing ]

        elif self.align > 0: # right
            return [ ' ' * (self.w - len(line)) + line for line in showing ]

        else: # center
            view = [ ]
            lpad = True

            for line in showing:
                diff = self.w - len(line)
                if diff:
                    left = ' ' * (diff // 2)
                    right = left
                    if diff % 2:
                        if lpad: left += ' '
                        else: right += ' '
                        lpad = not lpad

                    view.append(left + line + right)

                else:
                    view.append(line)

            return view




from .backend import core

from threading import Thread, Event, current_thread


class Application:
    def __init__(self):
        self.go = False
        self.quit = Event()
        self.error = None

        self.core = core.Application(self)
        self.windows = [ ]
        self.mouse = (False, False, False) # mouse buttons (left, mid, right)
        self.moving = None
        self.sizing = None

    def exit(self):
        self.core.exit()
        if self.error:
            raise self.error

    def run(self):
        self.go = True

        self.inputs = Thread(target=self.wrap(self.core.inputs),
            name='inputs', daemon=True)
        self.inputs.start()

        try:
            self.on_run()

        except Quit:
            pass

    def stop(self, error=None):
        if error:
            if self.error:
                try: raise self.error
                except Exception as e:
                    try: raise error
                    except Exception as e:
                        self.error = e

            else:
                self.error = error

        self.go = False
        self.quit.set()

    def wrap(self, fn):
        def wrapped():
            try: fn()
            except Exception as e:
                try: raise ThreadException() from e
                except Exception as e: self.stop(e)

        return wrapped

    def sleep(self, secs):
        if self.quit.wait(timeout=secs):
            raise Quit()

    @property
    def w(self):
        return self.core.w

    @property
    def h(self):
        return self.core.h

    def on_run(self):
        self.inputs.join()

    def on_key(self, key):
        if key == 27: # escape
            self.stop()

    def on_mouse(self, x, y, left, mid, right, scroll):
        if self.moving:
            win, mx, my = self.moving
            if not left:
                self.moving = None

            if x != mx or y != my:
                win.move_by(x - mx, y - my)
                if self.moving:
                    self.moving = (win, x, y)

        elif self.sizing:
            win, sxt, syt = self.sizing
            if not left:
                self.sizing = None

            if (sxt and x != sxt[0]) or (syt and y != syt[0]):
                mdx = 0
                mdy = 0
                sdx = 0
                sdy = 0

                if sxt:
                    if sxt[1]:
                        mdx = x - sxt[0]
                        sdx = -mdx
                    else:
                        sdx = x - sxt[0]

                if syt:
                    if syt[1]:
                        mdy = y - syt[0]
                        sdy = -mdy
                    else:
                        sdy = y - syt[0]

                if mdx or mdy:
                    win.move_by(mdx, mdy)

                win.size_by(sdx, sdy)

                if self.sizing:
                    self.sizing = (win,
                        (x, sxt[1]) if sxt else None,
                        (y, syt[1]) if syt else None)

        elif left and not self.mouse[0]: # left button pushed
            for win in self.windows:
                if x >= win.x and x < win.x + win.w \
                        and y >= win.y and y < win.y + win.h:
                    sx = (x, True) if x == win.x else \
                         (x, False) if x == win.x + win.w -1 else \
                         None
                    sy = (y, True) if y == win.y else \
                         (y, False) if y == win.y + win.h -1 else \
                         None

                    if sx or sy:
                        self.sizing = (win, sx, sy)

                    else:
                        self.moving = (win, x, y)

                    break

        elif scroll:
            for win in self.windows:
                if x >= win.x and x < win.x + win.w \
                        and y >= win.y and y < win.y + win.h:
                    if scroll < 0: win.text.down()
                    else: win.text.up()
                    break

        self.mouse = (left, mid, right)

        # FIXME this is just for testing (assumes we have at least one window):
        #self.windows[0].text.print(f'{x},{y}: {left},{mid},{right},{scroll}')


class Quit(Exception):
    pass


class ThreadException(Exception):
    def __init__(self):
        t = current_thread()
        super().__init__('from thread ' \
            + f'"{t.name}"' if t.name else f'{t.ident}')




from . import *


def run():
    app = Application()
    try:
        win = Window(app, 30, 5, 42, 9)
        win.text.word_wrap()
        win.text.top()
        win.text.print('ESC to quit.')
        win.text.print('Click drag and scroll to test mouse.')
        win.text.print()
        win.text.print("Here's something.", enter=False)
        win.text.print(' And this continues the previous print and also wraps.')
        for i in range(1, 11):
            win.text.print()
            win.text.print(i, enter=False)
            win.text.print('. This is a line of text that goes and continues' \
                + ' and keeps going some more just to see how it all works.')

        def on_run():
            while app.go:
                app.sleep(1)
                win.text.center()
                app.sleep(1)
                win.text.right()
                app.sleep(1)
                win.text.left()
        app.on_run = on_run

        #raise RuntimeError('__main__ test')
        app.run()

    finally:
        app.exit()


if __name__ == '__main__':
    run()




import win32con
import win32file
from win32console import *

from time import sleep


class Application:
    def __init__(self, shell):
        self.shell = shell

        self.console = PyConsoleScreenBufferType(win32file.CreateFile(
            "CONIN$",
            win32con.GENERIC_READ | win32con.GENERIC_WRITE,
            win32con.FILE_SHARE_READ,
            None,
            win32con.OPEN_EXISTING,
            0,
            0
        ))
        self.console.SetConsoleMode(ENABLE_WINDOW_INPUT | ENABLE_MOUSE_INPUT)
        self.buffer = CreateConsoleScreenBuffer()
        self.buffer.SetConsoleActiveScreenBuffer()

        size, _ = self.buffer.GetConsoleCursorInfo()
        self.buffer.SetConsoleCursorInfo(size, False)

        self.mpos = (0, 0) # mouse position

    def exit(self):
        size, _ = self.buffer.GetConsoleCursorInfo()
        self.buffer.SetConsoleCursorInfo(size, True)

        self.buffer.Close()

    def inputs(self):
        while self.shell.go:
            # poll for events before blocking and getting them
            # because the block prevents output from showing up
            while self.console.GetNumberOfConsoleInputEvents() < 1:
                sleep(0.01)

            for input in self.console.ReadConsoleInput(1):
                if input.EventType == KEY_EVENT:
                    if input.KeyDown:
                        if input.Char == '\0':
                            pass # FIXME - virtual key

                        elif input.Char == '\x03': # ctrl-c
                            raise KeyboardInterrupt()

                        else:
                            self.shell.on_key(ord(input.Char))

                elif input.EventType == MOUSE_EVENT:
                    flags = input.EventFlags # 1=push, 2=click, 4=scroll
                    pos = input.MousePosition
                    btn = input.ButtonState
                    scroll = -1 if btn & 4287102976 == 4287102976 else \
                              1 if btn & 7864320 == 7864320 else 0

                    if not scroll: # mouse position not reliable while scrolling
                        self.mpos = (pos.X, pos.Y)

                    self.shell.on_mouse(self.mpos[0], self.mpos[1],
                        True if btn & 1 == 1 else False,
                        True if btn & 4 == 4 else False,
                        True if btn & 2 == 2 else False,
                        scroll
                    )

    def write(self, x, y, txt):
        self.buffer.WriteConsoleOutputCharacter(txt, PyCOORDType(x, y))
        #dummy = PyINPUT_RECORDType(KEY_EVENT)
        #dummy.Char = '\0'
        #self.console.WriteConsoleInput([ dummy ])

    def clear(self):
        self.buffer.FillConsoleOutputCharacter(' ', self.w * self.h,
            PyCOORDType(0, 0))

    @property
    def w(self):
        return self.buffer.GetConsoleScreenBufferInfo()["Size"].X

    @property
    def h(self):
        return self.buffer.GetConsoleScreenBufferInfo()["Size"].Y


class Window:
    def __init__(self, app, x, y, w, h):
        self.app = app
        assert w >= 2 and h >= 2
        assert x >= 0 and x + w <= self.app.w
        assert y >= 0 and y + h <= self.app.h

        self.x = x
        self.y = y
        self.w = w
        self.h = h

        self.view = [ ]
        self._border_()

    def write(self, x, y, txt):
        assert x >= 0 and x + len(txt) <= self.w
        assert y >= 0 and y < self.h

        self.app.write(self.x + x, self.y + y, txt)

        while y >= len(self.view):
            self.view.append('')

        line = self.view[y]
        if len(line) < x:
            self.view[y] = line + ' ' * (x - len(line)) + txt
        else:
            self.view[y] = line[:x] + txt + line[x+len(txt):]

    def write_all(self, x, y, lines):
        assert x >= 0 and x < self.w
        assert y >= 0 and y < self.h
        assert y + len(lines) < self.h

        for line in lines:
            self.write(x, y, line)
            y += 1

    def set_pos(self, x, y):
        assert x >= 0 and x + self.w <= self.app.w
        assert y >= 0 and y + self.h <= self.app.h
        if x == self.x and y == self.y: return

        self.app.clear()
        self.x = x
        self.y = y
        self._review_()

    def set_size(self, w, h):
        assert w >= 2 and self.x + w <= self.app.w
        assert h >= 2 and self.y + h <= self.app.h
        if w == self.w and h == self.h: return

        self.app.clear()
        self.w = w
        self.h = h
        self._border_()
        self._review_()

    def _border_(self):
        self.write(0, 0, '\u250c')
        self.write(1, 0, '\u2500' * (self.w - 2))
        self.write(self.w - 1, 0, '\u2510')

        for y in range(1, self.h):
            self.write(0, y, '\u2502')
            self.write(self.w - 1, y, '\u2502')

        self.write(0, self.h - 1, '\u2514')
        self.write(1, self.h - 1, '\u2500' * (self.w - 2))
        self.write(self.w - 1, self.h - 1, '\u2518')

    def _review_(self):
        for y in range(self.h):
            if y >= len(self.view): break
            self.app.write(self.x, self.y + y, self.view[y][:self.w])




import platform
import os


windows = platform.system == 'Windows' or os.name == 'nt'
linux = platform.system == 'Linux'
macos = platform.system == 'Darwin'


if windows:
    from . import win32 as core

else:
    from . import curses as core




import curses


class Application:
    def __init__(self, shell):
        self.shell = shell

        self.screen = curses.initscr()
        curses.flushinp()
        curses.curs_set(0)
        curses.noecho()
        curses.cbreak()
        self.screen.keypad(True)
        curses.set_escdelay(222)
        curses.mousemask(curses.ALL_MOUSE_EVENTS | curses.REPORT_MOUSE_POSITION)
        curses.mouseinterval(0)

        self.mbtns = [False, False, False] # mouse buttons (left, mid, right)

    def exit(self):
        self.screen.keypad(False)
        curses.nocbreak()
        curses.echo()
        curses.curs_set(1)
        curses.flushinp()
        curses.endwin()

    def inputs(self):
        while self.shell.go:
            key = self.screen.getch()

            if key == curses.KEY_MOUSE:
                _, x, y, _, btn = curses.getmouse()
                mbtns = self.mbtns

                if btn & curses.BUTTON1_PRESSED:
                    mbtns[0] = True
                elif btn & curses.BUTTON1_RELEASED:
                    mbtns[0] = False

                if btn & curses.BUTTON2_PRESSED:
                    mbtns[1] = True
                elif btn & curses.BUTTON2_RELEASED:
                    mbtns[1] = False

                if btn & curses.BUTTON3_PRESSED:
                    mbtns[2] = True
                elif btn & curses.BUTTON3_RELEASED:
                    mbtns[2] = False

                if btn & curses.BUTTON1_CLICKED \
                        or btn & curses.BUTTON1_DOUBLE_CLICKED:
                    self.shell.on_mouse(x, y, True, mbtns[1], mbtns[2], 0)
                    self.shell.on_mouse(x, y, False, mbtns[1], mbtns[2], 0)

                    if btn & curses.BUTTON1_DOUBLE_CLICKED:
                        self.shell.on_mouse(x, y, True, mbtns[1], mbtns[2], 0)
                        self.shell.on_mouse(x, y, False, mbtns[1], mbtns[2], 0)

                elif btn & curses.BUTTON2_CLICKED \
                        or btn & curses.BUTTON2_DOUBLE_CLICKED:
                    self.shell.on_mouse(x, y, mbtns[0], True, mbtns[2], 0)
                    self.shell.on_mouse(x, y, mbtns[0], False, mbtns[2], 0)

                    if btn & curses.BUTTON2_DOUBLE_CLICKED:
                        self.shell.on_mouse(x, y, mbtns[0], True, mbtns[2], 0)
                        self.shell.on_mouse(x, y, mbtns[0], False, mbtns[2], 0)

                elif btn & curses.BUTTON3_CLICKED \
                        or btn & curses.BUTTON3_DOUBLE_CLICKED:
                    self.shell.on_mouse(x, y, mbtns[0], mbtns[1], True, 0)
                    self.shell.on_mouse(x, y, mbtns[0], mbtns[1], False, 0)

                    if btn & curses.BUTTON3_DOUBLE_CLICKED:
                        self.shell.on_mouse(x, y, mbtns[0], mbtns[1], True, 0)
                        self.shell.on_mouse(x, y, mbtns[0], mbtns[1], False, 0)

                elif btn & curses.BUTTON4_PRESSED:
                    self.shell.on_mouse(x, y, mbtns[0], mbtns[1], mbtns[2], 1)

                elif btn & 2097152:
                    self.shell.on_mouse(x, y, mbtns[0], mbtns[1], mbtns[2], -1)

                else:
                    self.shell.on_mouse(x, y, mbtns[0], mbtns[1], mbtns[2], 0)

            else:
                #raise RuntimeError('just a test')
                self.shell.on_key(key)

    @property
    def w(self):
        return curses.COLS

    @property
    def h(self):
        return curses.LINES


class Window:
    def __init__(self, app, x, y, w, h):
        self.app = app
        assert w >= 2 and h >= 2
        assert x >= 0 and x + w <= self.app.w
        assert y >= 0 and y + h <= self.app.h

        win = curses.newwin(h, w, y, x)
        win.keypad(True)
        win.box()
        self.win = win

        self.app.screen.refresh()
        self.win.refresh()

    @property
    def x(self):
        return self.win.getbegyx()[1]

    @property
    def y(self):
        return self.win.getbegyx()[0]

    @property
    def w(self):
        return self.win.getmaxyx()[1]

    @property
    def h(self):
        return self.win.getmaxyx()[0]

    def write(self, x, y, txt):
        assert x >= 0 and x + len(txt) <= self.w
        assert y >= 0 and y < self.h

        self.win.addstr(y, x, txt)
        self.win.refresh()

    def write_all(self, x, y, lines):
        assert x >= 0 and x < self.w
        assert y >= 0 and y < self.h
        assert y + len(lines) < self.h

        for line in lines:
            assert x + len(line) <= self.w
            self.win.addstr(y, x, line)
            y += 1

        self.win.refresh()

    def set_pos(self, x, y):
        assert x >= 0 and x + self.w <= self.app.w
        assert y >= 0 and y + self.h <= self.app.h

        self.app.screen.clear()
        self.app.screen.refresh()
        self.win.mvwin(y, x)
        self.win.refresh()

    def set_size(self, w, h):
        assert w >= 2 and self.x + w <= self.app.w
        assert h >= 2 and self.y + h <= self.app.h

        self.app.screen.clear()
        self.app.screen.refresh()
        self.win.resize(h, w)
        self.win.box()
        self.win.refresh()

